@@ -1,50 +1,50 @@
 use std::borrow::Borrow;
 use std::cmp;
+use std::convert::TryFrom;
 use std::fmt;
 use std::hash::{Hash, Hasher};
-use std::ptr;
+use std::mem::MaybeUninit;
 use std::ops::{Deref, DerefMut};
+use std::ptr;
+use std::slice;
 use std::str;
 use std::str::FromStr;
 use std::str::Utf8Error;
-use std::slice;
 
-use crate::array::Array;
-use crate::array::Index;
 use crate::CapacityError;
+use crate::LenUint;
 use crate::char::encode_utf8;
 
 #[cfg(feature="serde")]
 use serde::{Serialize, Deserialize, Serializer, Deserializer};
 
-use super::MaybeUninit as MaybeUninitCopy;
 
 /// A string with a fixed capacity.
 ///
 /// The `ArrayString` is a string backed by a fixed size array. It keeps track
-/// of its length.
+/// of its length, and is parameterized by `CAP` for the maximum capacity.
+///
+/// `CAP` is of type `usize` but is range limited to `u32::MAX`; attempting to create larger
+/// arrayvecs with larger capacity will panic.
 ///
 /// The string is a contiguous value that you can store directly on the stack
 /// if needed.
 #[derive(Copy)]
-pub struct ArrayString<A>
-    where A: Array<Item=u8> + Copy
-{
-    xs: MaybeUninitCopy<A>,
-    len: A::Index,
+pub struct ArrayString<const CAP: usize> {
+    // the `len` first elements of the array are initialized
+    xs: [MaybeUninit<u8>; CAP],
+    len: LenUint,
 }
 
-impl<A> Default for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> Default for ArrayString<CAP>
 {
     /// Return an empty `ArrayString`
-    fn default() -> ArrayString<A> {
+    fn default() -> ArrayString<CAP> {
         ArrayString::new()
     }
 }
 
-impl<A> ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> ArrayString<CAP>
 {
     /// Create a new empty `ArrayString`.
     ///
@@ -53,20 +53,35 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     ///
-    /// let mut string = ArrayString::<[_; 16]>::new();
+    /// let mut string = ArrayString::<16>::new();
     /// string.push_str("foo");
     /// assert_eq!(&string[..], "foo");
     /// assert_eq!(string.capacity(), 16);
     /// ```
-    pub fn new() -> ArrayString<A> {
+    #[cfg(not(feature="unstable-const-fn"))]
+    pub fn new() -> ArrayString<CAP> {
+        assert_capacity_limit!(CAP);
         unsafe {
-            ArrayString {
-                xs: MaybeUninitCopy::uninitialized(),
-                len: Index::from(0),
-            }
+            ArrayString { xs: MaybeUninit::uninit().assume_init(), len: 0 }
         }
     }
 
+    #[cfg(feature="unstable-const-fn")]
+    pub const fn new() -> ArrayString<CAP> {
+        assert_capacity_limit!(CAP);
+        unsafe {
+            ArrayString { xs: MaybeUninit::uninit().assume_init(), len: 0 }
+        }
+    }
+
+    /// Return the length of the string.
+    #[inline]
+    pub fn len(&self) -> usize { self.len as usize }
+
+    /// Returns whether the string is empty.
+    #[inline]
+    pub fn is_empty(&self) -> bool { self.len() == 0 }
+
     /// Create a new `ArrayString` from a `str`.
     ///
     /// Capacity is inferred from the type parameter.
@@ -76,7 +91,7 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     ///
-    /// let mut string = ArrayString::<[_; 3]>::from("foo").unwrap();
+    /// let mut string = ArrayString::<3>::from("foo").unwrap();
     /// assert_eq!(&string[..], "foo");
     /// assert_eq!(string.len(), 3);
     /// assert_eq!(string.capacity(), 3);
@@ -96,13 +111,16 @@ impl<A> ArrayString<A>
     ///
     /// let string = ArrayString::from_byte_string(b"hello world").unwrap();
     /// ```
-    pub fn from_byte_string(b: &A) -> Result<Self, Utf8Error> {
-        let len = str::from_utf8(b.as_slice())?.len();
-        debug_assert_eq!(len, A::CAPACITY);
-        Ok(ArrayString {
-            xs: MaybeUninitCopy::from(*b),
-            len: Index::from(A::CAPACITY),
-        })
+    pub fn from_byte_string(b: &[u8; CAP]) -> Result<Self, Utf8Error> {
+        let len = str::from_utf8(b)?.len();
+        debug_assert_eq!(len, CAP);
+        let mut vec = Self::new();
+        unsafe {
+            (b as *const [u8; CAP] as *const [MaybeUninit<u8>; CAP])
+                .copy_to_nonoverlapping(&mut vec.xs as *mut [MaybeUninit<u8>; CAP], 1);
+            vec.set_len(CAP);
+        }
+        Ok(vec)
     }
 
     /// Return the capacity of the `ArrayString`.
@@ -110,18 +128,18 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     ///
-    /// let string = ArrayString::<[_; 3]>::new();
+    /// let string = ArrayString::<3>::new();
     /// assert_eq!(string.capacity(), 3);
     /// ```
-    #[inline]
-    pub fn capacity(&self) -> usize { A::CAPACITY }
+    #[inline(always)]
+    pub fn capacity(&self) -> usize { CAP }
 
     /// Return if the `ArrayString` is completely filled.
     ///
     /// ```
     /// use arrayvec::ArrayString;
     ///
-    /// let mut string = ArrayString::<[_; 1]>::new();
+    /// let mut string = ArrayString::<1>::new();
     /// assert!(!string.is_full());
     /// string.push_str("A");
     /// assert!(string.is_full());
@@ -135,7 +153,7 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     ///
-    /// let mut string = ArrayString::<[_; 2]>::new();
+    /// let mut string = ArrayString::<2>::new();
     ///
     /// string.push('a');
     /// string.push('b');
@@ -155,7 +173,7 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     ///
-    /// let mut string = ArrayString::<[_; 2]>::new();
+    /// let mut string = ArrayString::<2>::new();
     ///
     /// string.try_push('a').unwrap();
     /// string.try_push('b').unwrap();
@@ -167,7 +185,9 @@ impl<A> ArrayString<A>
     pub fn try_push(&mut self, c: char) -> Result<(), CapacityError<char>> {
         let len = self.len();
         unsafe {
-            match encode_utf8(c, &mut self.raw_mut_bytes()[len..]) {
+            let ptr = self.as_mut_ptr().add(len);
+            let remaining_cap = self.capacity() - len;
+            match encode_utf8(c, ptr, remaining_cap) {
                 Ok(n) => {
                     self.set_len(len + n);
                     Ok(())
@@ -184,7 +204,7 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     ///
-    /// let mut string = ArrayString::<[_; 2]>::new();
+    /// let mut string = ArrayString::<2>::new();
     ///
     /// string.push_str("a");
     /// string.push_str("d");
@@ -204,7 +224,7 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     ///
-    /// let mut string = ArrayString::<[_; 2]>::new();
+    /// let mut string = ArrayString::<2>::new();
     ///
     /// string.try_push_str("a").unwrap();
     /// let overflow1 = string.try_push_str("bc");
@@ -220,7 +240,7 @@ impl<A> ArrayString<A>
             return Err(CapacityError::new(s));
         }
         unsafe {
-            let dst = self.xs.ptr_mut().offset(self.len() as isize);
+            let dst = self.as_mut_ptr().add(self.len());
             let src = s.as_ptr();
             ptr::copy_nonoverlapping(src, dst, s.len());
             let newl = self.len() + s.len();
@@ -236,7 +256,7 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     /// 
-    /// let mut s = ArrayString::<[_; 3]>::from("foo").unwrap();
+    /// let mut s = ArrayString::<3>::from("foo").unwrap();
     ///
     /// assert_eq!(s.pop(), Some('o'));
     /// assert_eq!(s.pop(), Some('o'));
@@ -244,7 +264,6 @@ impl<A> ArrayString<A>
     ///
     /// assert_eq!(s.pop(), None);
     /// ```
-    #[inline]
     pub fn pop(&mut self) -> Option<char> {
         let ch = match self.chars().rev().next() {
             Some(ch) => ch,
@@ -267,13 +286,12 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     ///
-    /// let mut string = ArrayString::<[_; 6]>::from("foobar").unwrap();
+    /// let mut string = ArrayString::<6>::from("foobar").unwrap();
     /// string.truncate(3);
     /// assert_eq!(&string[..], "foo");
     /// string.truncate(4);
     /// assert_eq!(&string[..], "foo");
     /// ```
-    #[inline]
     pub fn truncate(&mut self, new_len: usize) {
         if new_len <= self.len() {
             assert!(self.is_char_boundary(new_len));
@@ -298,13 +316,12 @@ impl<A> ArrayString<A>
     /// ```
     /// use arrayvec::ArrayString;
     /// 
-    /// let mut s = ArrayString::<[_; 3]>::from("foo").unwrap();
+    /// let mut s = ArrayString::<3>::from("foo").unwrap();
     ///
     /// assert_eq!(s.remove(0), 'f');
     /// assert_eq!(s.remove(1), 'o');
     /// assert_eq!(s.remove(0), 'o');
     /// ```
-    #[inline]
     pub fn remove(&mut self, idx: usize) -> char {
         let ch = match self[idx..].chars().next() {
             Some(ch) => ch,
@@ -314,8 +331,8 @@ impl<A> ArrayString<A>
         let next = idx + ch.len_utf8();
         let len = self.len();
         unsafe {
-            ptr::copy(self.xs.ptr().offset(next as isize),
-                      self.xs.ptr_mut().offset(idx as isize),
+            ptr::copy(self.as_ptr().add(next),
+                      self.as_mut_ptr().add(idx),
                       len - next);
             self.set_len(len - (next - idx));
         }
@@ -336,10 +353,10 @@ impl<A> ArrayString<A>
     ///
     /// This method uses *debug assertions* to check the validity of `length`
     /// and may use other debug assertions.
-    #[inline]
     pub unsafe fn set_len(&mut self, length: usize) {
+        // type invariant that capacity always fits in LenUint
         debug_assert!(length <= self.capacity());
-        self.len = Index::from(length);
+        self.len = length as LenUint;
     }
 
     /// Return a string slice of the whole `ArrayString`.
@@ -347,100 +364,92 @@ impl<A> ArrayString<A>
         self
     }
 
-    /// Return a mutable slice of the whole string’s buffer
-    unsafe fn raw_mut_bytes(&mut self) -> &mut [u8] {
-        slice::from_raw_parts_mut(self.xs.ptr_mut(), self.capacity())
+    fn as_ptr(&self) -> *const u8 {
+        self.xs.as_ptr() as *const u8
+    }
+
+    fn as_mut_ptr(&mut self) -> *mut u8 {
+        self.xs.as_mut_ptr() as *mut u8
     }
 }
 
-impl<A> Deref for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> Deref for ArrayString<CAP>
 {
     type Target = str;
     #[inline]
     fn deref(&self) -> &str {
         unsafe {
-            let sl = slice::from_raw_parts(self.xs.ptr(), self.len.to_usize());
+            let sl = slice::from_raw_parts(self.as_ptr(), self.len());
             str::from_utf8_unchecked(sl)
         }
     }
 }
 
-impl<A> DerefMut for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> DerefMut for ArrayString<CAP>
 {
     #[inline]
     fn deref_mut(&mut self) -> &mut str {
         unsafe {
-            let sl = slice::from_raw_parts_mut(self.xs.ptr_mut(), self.len.to_usize());
+            let len = self.len();
+            let sl = slice::from_raw_parts_mut(self.as_mut_ptr(), len);
             str::from_utf8_unchecked_mut(sl)
         }
     }
 }
 
-impl<A> PartialEq for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> PartialEq for ArrayString<CAP>
 {
     fn eq(&self, rhs: &Self) -> bool {
         **self == **rhs
     }
 }
 
-impl<A> PartialEq<str> for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> PartialEq<str> for ArrayString<CAP>
 {
     fn eq(&self, rhs: &str) -> bool {
         &**self == rhs
     }
 }
 
-impl<A> PartialEq<ArrayString<A>> for str
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> PartialEq<ArrayString<CAP>> for str
 {
-    fn eq(&self, rhs: &ArrayString<A>) -> bool {
+    fn eq(&self, rhs: &ArrayString<CAP>) -> bool {
         self == &**rhs
     }
 }
 
-impl<A> Eq for ArrayString<A> 
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> Eq for ArrayString<CAP> 
 { }
 
-impl<A> Hash for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> Hash for ArrayString<CAP>
 {
     fn hash<H: Hasher>(&self, h: &mut H) {
         (**self).hash(h)
     }
 }
 
-impl<A> Borrow<str> for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> Borrow<str> for ArrayString<CAP>
 {
     fn borrow(&self) -> &str { self }
 }
 
-impl<A> AsRef<str> for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> AsRef<str> for ArrayString<CAP>
 {
     fn as_ref(&self) -> &str { self }
 }
 
-impl<A> fmt::Debug for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> fmt::Debug for ArrayString<CAP>
 {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result { (**self).fmt(f) }
 }
 
-impl<A> fmt::Display for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> fmt::Display for ArrayString<CAP>
 {
     fn fmt(&self, f: &mut fmt::Formatter) -> fmt::Result { (**self).fmt(f) }
 }
 
 /// `Write` appends written data to the end of the string.
-impl<A> fmt::Write for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> fmt::Write for ArrayString<CAP>
 {
     fn write_char(&mut self, c: char) -> fmt::Result {
         self.try_push(c).map_err(|_| fmt::Error)
@@ -451,10 +460,9 @@ impl<A> fmt::Write for ArrayString<A>
     }
 }
 
-impl<A> Clone for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> Clone for ArrayString<CAP>
 {
-    fn clone(&self) -> ArrayString<A> {
+    fn clone(&self) -> ArrayString<CAP> {
         *self
     }
     fn clone_from(&mut self, rhs: &Self) {
@@ -464,8 +472,7 @@ impl<A> Clone for ArrayString<A>
     }
 }
 
-impl<A> PartialOrd for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> PartialOrd for ArrayString<CAP>
 {
     fn partial_cmp(&self, rhs: &Self) -> Option<cmp::Ordering> {
         (**self).partial_cmp(&**rhs)
@@ -476,8 +483,7 @@ impl<A> PartialOrd for ArrayString<A>
     fn ge(&self, rhs: &Self) -> bool { **self >= **rhs }
 }
 
-impl<A> PartialOrd<str> for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> PartialOrd<str> for ArrayString<CAP>
 {
     fn partial_cmp(&self, rhs: &str) -> Option<cmp::Ordering> {
         (**self).partial_cmp(rhs)
@@ -488,28 +494,25 @@ impl<A> PartialOrd<str> for ArrayString<A>
     fn ge(&self, rhs: &str) -> bool { &**self >= rhs }
 }
 
-impl<A> PartialOrd<ArrayString<A>> for str
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> PartialOrd<ArrayString<CAP>> for str
 {
-    fn partial_cmp(&self, rhs: &ArrayString<A>) -> Option<cmp::Ordering> {
+    fn partial_cmp(&self, rhs: &ArrayString<CAP>) -> Option<cmp::Ordering> {
         self.partial_cmp(&**rhs)
     }
-    fn lt(&self, rhs: &ArrayString<A>) -> bool { self < &**rhs }
-    fn le(&self, rhs: &ArrayString<A>) -> bool { self <= &**rhs }
-    fn gt(&self, rhs: &ArrayString<A>) -> bool { self > &**rhs }
-    fn ge(&self, rhs: &ArrayString<A>) -> bool { self >= &**rhs }
+    fn lt(&self, rhs: &ArrayString<CAP>) -> bool { self < &**rhs }
+    fn le(&self, rhs: &ArrayString<CAP>) -> bool { self <= &**rhs }
+    fn gt(&self, rhs: &ArrayString<CAP>) -> bool { self > &**rhs }
+    fn ge(&self, rhs: &ArrayString<CAP>) -> bool { self >= &**rhs }
 }
 
-impl<A> Ord for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> Ord for ArrayString<CAP>
 {
     fn cmp(&self, rhs: &Self) -> cmp::Ordering {
         (**self).cmp(&**rhs)
     }
 }
 
-impl<A> FromStr for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> FromStr for ArrayString<CAP>
 {
     type Err = CapacityError;
 
@@ -520,8 +523,7 @@ impl<A> FromStr for ArrayString<A>
 
 #[cfg(feature="serde")]
 /// Requires crate feature `"serde"`
-impl<A> Serialize for ArrayString<A>
-    where A: Array<Item=u8> + Copy
+impl<const CAP: usize> Serialize for ArrayString<CAP>
 {
     fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
         where S: Serializer
@@ -532,8 +534,7 @@ impl<A> Serialize for ArrayString<A>
 
 #[cfg(feature="serde")]
 /// Requires crate feature `"serde"`
-impl<'de, A> Deserialize<'de> for ArrayString<A> 
-    where A: Array<Item=u8> + Copy
+impl<'de, const CAP: usize> Deserialize<'de> for ArrayString<CAP> 
 {
     fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
         where D: Deserializer<'de>
@@ -541,13 +542,13 @@ impl<'de, A> Deserialize<'de> for ArrayString<A>
         use serde::de::{self, Visitor};
         use std::marker::PhantomData;
 
-        struct ArrayStringVisitor<A: Array<Item=u8>>(PhantomData<A>);
+        struct ArrayStringVisitor<const CAP: usize>(PhantomData<[u8; CAP]>);
 
-        impl<'de, A: Copy + Array<Item=u8>> Visitor<'de> for ArrayStringVisitor<A> {
-            type Value = ArrayString<A>;
+        impl<'de, const CAP: usize> Visitor<'de> for ArrayStringVisitor<CAP> {
+            type Value = ArrayString<CAP>;
 
             fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
-                write!(formatter, "a string no more than {} bytes long", A::CAPACITY)
+                write!(formatter, "a string no more than {} bytes long", CAP)
             }
 
             fn visit_str<E>(self, v: &str) -> Result<Self::Value, E>
@@ -565,6 +566,29 @@ impl<'de, A> Deserialize<'de> for ArrayString<A>
             }
         }
 
-        deserializer.deserialize_str(ArrayStringVisitor::<A>(PhantomData))
+        deserializer.deserialize_str(ArrayStringVisitor(PhantomData))
+    }
+}
+
+impl<'a, const CAP: usize> TryFrom<&'a str> for ArrayString<CAP>
+{
+    type Error = CapacityError<&'a str>;
+
+    fn try_from(f: &'a str) -> Result<Self, Self::Error> {
+        let mut v = Self::new();
+        v.try_push_str(f)?;
+        Ok(v)
+    }
+}
+
+impl<'a, const CAP: usize> TryFrom<fmt::Arguments<'a>> for ArrayString<CAP>
+{
+    type Error = CapacityError<fmt::Error>;
+
+    fn try_from(f: fmt::Arguments<'a>) -> Result<Self, Self::Error> {
+        use fmt::Write;
+        let mut v = Self::new();
+        v.write_fmt(f).map_err(|e| CapacityError::new(e))?;
+        Ok(v)
     }
 }
