@@ -13,7 +13,7 @@ use std::collections::HashMap;
 fn test_simple() {
     use std::ops::Add;
 
-    let mut vec: ArrayVec<[Vec<i32>; 3]> = ArrayVec::new();
+    let mut vec: ArrayVec<Vec<i32>,  3> = ArrayVec::new();
 
     vec.push(vec![1, 2, 3, 4]);
     vec.push(vec![10]);
@@ -27,10 +27,57 @@ fn test_simple() {
     assert_eq!(sum_len, 8);
 }
 
+#[test]
+fn test_capacity_left() {
+    let mut vec: ArrayVec<usize,  4> = ArrayVec::new();
+    assert_eq!(vec.remaining_capacity(), 4);
+    vec.push(1);
+    assert_eq!(vec.remaining_capacity(), 3);
+    vec.push(2);
+    assert_eq!(vec.remaining_capacity(), 2);
+    vec.push(3);
+    assert_eq!(vec.remaining_capacity(), 1);
+    vec.push(4);
+    assert_eq!(vec.remaining_capacity(), 0);
+}
+
+#[test]
+fn test_extend_from_slice() {
+    let mut vec: ArrayVec<usize,  10> = ArrayVec::new();
+
+    vec.try_extend_from_slice(&[1, 2, 3]).unwrap();
+    assert_eq!(vec.len(), 3);
+    assert_eq!(&vec[..], &[1, 2, 3]);
+    assert_eq!(vec.pop(), Some(3));
+    assert_eq!(&vec[..], &[1, 2]);
+}
+
+#[test]
+fn test_extend_from_slice_error() {
+    let mut vec: ArrayVec<usize,  10> = ArrayVec::new();
+
+    vec.try_extend_from_slice(&[1, 2, 3]).unwrap();
+    let res = vec.try_extend_from_slice(&[0; 8]);
+    assert_matches!(res, Err(_));
+
+    let mut vec: ArrayVec<usize,  0> = ArrayVec::new();
+    let res = vec.try_extend_from_slice(&[0; 1]);
+    assert_matches!(res, Err(_));
+}
+
+#[test]
+fn test_try_from_slice_error() {
+    use arrayvec::ArrayVec;
+    use std::convert::TryInto as _;
+
+    let res: Result<ArrayVec<_,  2>, _> = (&[1, 2, 3] as &[_]).try_into();
+    assert_matches!(res, Err(_));
+}
+
 #[test]
 fn test_u16_index() {
     const N: usize = 4096;
-    let mut vec: ArrayVec<[_; N]> = ArrayVec::new();
+    let mut vec: ArrayVec<_,  N> = ArrayVec::new();
     for _ in 0..N {
         assert!(vec.try_push(1u8).is_ok());
     }
@@ -55,6 +102,7 @@ fn test_drop() {
 
     let flag = &Cell::new(0);
 
+    #[derive(Clone)]
     struct Bump<'a>(&'a Cell<i32>);
 
     impl<'a> Drop for Bump<'a> {
@@ -65,7 +113,7 @@ fn test_drop() {
     }
 
     {
-        let mut array = ArrayVec::<[Bump; 128]>::new();
+        let mut array = ArrayVec::<Bump,  128>::new();
         array.push(Bump(flag));
         array.push(Bump(flag));
     }
@@ -75,7 +123,7 @@ fn test_drop() {
     flag.set(0);
 
     {
-        let mut array = ArrayVec::<[_; 3]>::new();
+        let mut array = ArrayVec::<_,  3>::new();
         array.push(vec![Bump(flag)]);
         array.push(vec![Bump(flag), Bump(flag)]);
         array.push(vec![]);
@@ -94,7 +142,7 @@ fn test_drop() {
     // test into_inner
     flag.set(0);
     {
-        let mut array = ArrayVec::<[_; 3]>::new();
+        let mut array = ArrayVec::<_,  3>::new();
         array.push(Bump(flag));
         array.push(Bump(flag));
         array.push(Bump(flag));
@@ -105,44 +153,167 @@ fn test_drop() {
         assert_eq!(flag.get(), 3);
     }
 
+    // test cloning into_iter
+    flag.set(0);
+    {
+        let mut array = ArrayVec::<_,  3>::new();
+        array.push(Bump(flag));
+        array.push(Bump(flag));
+        array.push(Bump(flag));
+        let mut iter = array.into_iter();
+        assert_eq!(flag.get(), 0);
+        iter.next();
+        assert_eq!(flag.get(), 1);
+        let clone = iter.clone();
+        assert_eq!(flag.get(), 1);
+        drop(clone);
+        assert_eq!(flag.get(), 3);
+        drop(iter);
+        assert_eq!(flag.get(), 5);
+    }
+}
+
+#[test]
+fn test_drop_panics() {
+    use std::cell::Cell;
+    use std::panic::catch_unwind;
+    use std::panic::AssertUnwindSafe;
+
+    let flag = &Cell::new(0);
+
+    struct Bump<'a>(&'a Cell<i32>);
+
+    // Panic in the first drop
+    impl<'a> Drop for Bump<'a> {
+        fn drop(&mut self) {
+            let n = self.0.get();
+            self.0.set(n + 1);
+            if n == 0 {
+                panic!("Panic in Bump's drop");
+            }
+        }
+    }
+    // check if rust is new enough
+    flag.set(0);
+    {
+        let array = vec![Bump(flag), Bump(flag)];
+        let res = catch_unwind(AssertUnwindSafe(|| {
+            drop(array);
+        }));
+        assert!(res.is_err());
+    }
+
+    if flag.get() != 2 {
+        println!("test_drop_panics: skip, this version of Rust doesn't continue in drop_in_place");
+        return;
+    }
+
+    flag.set(0);
+    {
+        let mut array = ArrayVec::<Bump,  128>::new();
+        array.push(Bump(flag));
+        array.push(Bump(flag));
+        array.push(Bump(flag));
+
+        let res = catch_unwind(AssertUnwindSafe(|| {
+            drop(array);
+        }));
+        assert!(res.is_err());
+    }
+    // Check that all the elements drop, even if the first drop panics.
+    assert_eq!(flag.get(), 3);
+
+
+    flag.set(0);
+    {
+        let mut array = ArrayVec::<Bump,  16>::new();
+        array.push(Bump(flag));
+        array.push(Bump(flag));
+        array.push(Bump(flag));
+        array.push(Bump(flag));
+        array.push(Bump(flag));
+
+        let i = 2;
+        let tail_len = array.len() - i;
+
+        let res = catch_unwind(AssertUnwindSafe(|| {
+            array.truncate(i);
+        }));
+        assert!(res.is_err());
+        // Check that all the tail elements drop, even if the first drop panics.
+        assert_eq!(flag.get(), tail_len as i32);
+    }
+
+
 }
 
 #[test]
 fn test_extend() {
     let mut range = 0..10;
 
-    let mut array: ArrayVec<[_; 5]> = range.by_ref().collect();
+    let mut array: ArrayVec<_,  5> = range.by_ref().take(5).collect();
     assert_eq!(&array[..], &[0, 1, 2, 3, 4]);
     assert_eq!(range.next(), Some(5));
 
-    array.extend(range.by_ref());
+    array.extend(range.by_ref().take(0));
     assert_eq!(range.next(), Some(6));
 
-    let mut array: ArrayVec<[_; 10]> = (0..3).collect();
+    let mut array: ArrayVec<_,  10> = (0..3).collect();
     assert_eq!(&array[..], &[0, 1, 2]);
     array.extend(3..5);
     assert_eq!(&array[..], &[0, 1, 2, 3, 4]);
 }
 
+#[should_panic]
+#[test]
+fn test_extend_capacity_panic_1() {
+    let mut range = 0..10;
+
+    let _: ArrayVec<_,  5> = range.by_ref().collect();
+}
+
+#[should_panic]
+#[test]
+fn test_extend_capacity_panic_2() {
+    let mut range = 0..10;
+
+    let mut array: ArrayVec<_,  5> = range.by_ref().take(5).collect();
+    assert_eq!(&array[..], &[0, 1, 2, 3, 4]);
+    assert_eq!(range.next(), Some(5));
+    array.extend(range.by_ref().take(1));
+}
+
 #[test]
 fn test_is_send_sync() {
-    let data = ArrayVec::<[Vec<i32>; 5]>::new();
-    &data as &Send;
-    &data as &Sync;
+    let data = ArrayVec::<Vec<i32>,  5>::new();
+    &data as &dyn Send;
+    &data as &dyn Sync;
 }
 
 #[test]
 fn test_compact_size() {
-    // Future rust will kill these drop flags!
-    // 4 elements size + 1 len + 1 enum tag + [1 drop flag]
-    type ByteArray = ArrayVec<[u8; 4]>;
+    // 4 bytes + padding + length
+    type ByteArray = ArrayVec<u8,  4>;
     println!("{}", mem::size_of::<ByteArray>());
-    assert!(mem::size_of::<ByteArray>() <= 8);
+    assert!(mem::size_of::<ByteArray>() <= 2 * mem::size_of::<u32>());
 
-    // 12 element size + 1 enum tag + 3 padding + 1 len + 1 drop flag + 2 padding
-    type QuadArray = ArrayVec<[u32; 3]>;
+    // just length
+    type EmptyArray = ArrayVec<u8,  0>;
+    println!("{}", mem::size_of::<EmptyArray>());
+    assert!(mem::size_of::<EmptyArray>() <= mem::size_of::<u32>());
+
+    // 3 elements + padding + length
+    type QuadArray = ArrayVec<u32,  3>;
     println!("{}", mem::size_of::<QuadArray>());
-    assert!(mem::size_of::<QuadArray>() <= 24);
+    assert!(mem::size_of::<QuadArray>() <= 4 * 4 + mem::size_of::<u32>());
+}
+
+#[test]
+fn test_still_works_with_option_arrayvec() {
+    type RefArray = ArrayVec<&'static i32,  2>;
+    let array = Some(RefArray::new());
+    assert!(array.is_some());
+    println!("{:?}", array);
 }
 
 #[test]
@@ -152,16 +323,39 @@ fn test_drain() {
     v.drain(0..7);
     assert_eq!(&v[..], &[]);
 
-    v.extend(0..);
+    v.extend(0..8);
     v.drain(1..4);
     assert_eq!(&v[..], &[0, 4, 5, 6, 7]);
-    let u: ArrayVec<[_; 3]> = v.drain(1..4).rev().collect();
+    let u: ArrayVec<_,  3> = v.drain(1..4).rev().collect();
     assert_eq!(&u[..], &[6, 5, 4]);
     assert_eq!(&v[..], &[0, 7]);
     v.drain(..);
     assert_eq!(&v[..], &[]);
 }
 
+#[test]
+fn test_drain_range_inclusive() {
+    let mut v = ArrayVec::from([0; 8]);
+    v.drain(0..=7);
+    assert_eq!(&v[..], &[]);
+
+    v.extend(0..8);
+    v.drain(1..=4);
+    assert_eq!(&v[..], &[0, 5, 6, 7]);
+    let u: ArrayVec<_,  3> = v.drain(1..=2).rev().collect();
+    assert_eq!(&u[..], &[6, 5]);
+    assert_eq!(&v[..], &[0, 7]);
+    v.drain(..);
+    assert_eq!(&v[..], &[]);
+}
+
+#[test]
+#[should_panic]
+fn test_drain_range_inclusive_oob() {
+    let mut v = ArrayVec::from([0; 0]);
+    v.drain(0..=0);
+}
+
 #[test]
 fn test_retain() {
     let mut v = ArrayVec::from([0; 8]);
@@ -198,7 +392,7 @@ fn test_drop_panic() {
         }
     }
 
-    let mut array = ArrayVec::<[DropPanic; 1]>::new();
+    let mut array = ArrayVec::<DropPanic,  1>::new();
     array.push(DropPanic);
 }
 
@@ -213,7 +407,7 @@ fn test_drop_panic_into_iter() {
         }
     }
 
-    let mut array = ArrayVec::<[DropPanic; 1]>::new();
+    let mut array = ArrayVec::<DropPanic,  1>::new();
     array.push(DropPanic);
     array.into_iter();
 }
@@ -223,7 +417,7 @@ fn test_insert() {
     let mut v = ArrayVec::from([]);
     assert_matches!(v.try_push(1), Err(_));
 
-    let mut v = ArrayVec::<[_; 3]>::new();
+    let mut v = ArrayVec::<_,  3>::new();
     v.insert(0, 0);
     v.insert(1, 1);
     //let ret1 = v.try_insert(3, 3);
@@ -252,7 +446,7 @@ fn test_into_inner_1() {
 
 #[test]
 fn test_into_inner_2() {
-    let mut v = ArrayVec::<[String; 4]>::new();
+    let mut v = ArrayVec::<String,  4>::new();
     v.push("a".into());
     v.push("b".into());
     v.push("c".into());
@@ -261,16 +455,17 @@ fn test_into_inner_2() {
 }
 
 #[test]
-fn test_into_inner_3_() {
-    let mut v = ArrayVec::<[i32; 4]>::new();
-    v.extend(1..);
+fn test_into_inner_3() {
+    let mut v = ArrayVec::<i32,  4>::new();
+    v.extend(1..=4);
     assert_eq!(v.into_inner().unwrap(), [1, 2, 3, 4]);
 }
 
+#[cfg(feature="std")]
 #[test]
 fn test_write() {
     use std::io::Write;
-    let mut v = ArrayVec::<[_; 8]>::new();
+    let mut v = ArrayVec::<_,  8>::new();
     write!(&mut v, "\x01\x02\x03").unwrap();
     assert_eq!(&v[..], &[1, 2, 3]);
     let r = v.write(&[9; 16]).unwrap();
@@ -280,16 +475,16 @@ fn test_write() {
 
 #[test]
 fn array_clone_from() {
-    let mut v = ArrayVec::<[_; 4]>::new();
+    let mut v = ArrayVec::<_,  4>::new();
     v.push(vec![1, 2]);
     v.push(vec![3, 4, 5]);
     v.push(vec![6]);
     let reference = v.to_vec();
-    let mut u = ArrayVec::<[_; 4]>::new();
+    let mut u = ArrayVec::<_,  4>::new();
     u.clone_from(&v);
     assert_eq!(&u, &reference[..]);
 
-    let mut t = ArrayVec::<[_; 4]>::new();
+    let mut t = ArrayVec::<_,  4>::new();
     t.push(vec![97]);
     t.push(vec![]);
     t.push(vec![5, 6, 2]);
@@ -301,12 +496,13 @@ fn array_clone_from() {
     assert_eq!(&t, &reference[..]);
 }
 
+#[cfg(feature="std")]
 #[test]
 fn test_string() {
     use std::error::Error;
 
     let text = "hello world";
-    let mut s = ArrayString::<[_; 16]>::new();
+    let mut s = ArrayString::<16>::new();
     s.try_push_str(text).unwrap();
     assert_eq!(&s, text);
     assert_eq!(text, &s);
@@ -316,7 +512,7 @@ fn test_string() {
     map.insert(s, 1);
     assert_eq!(map[text], 1);
 
-    let mut t = ArrayString::<[_; 2]>::new();
+    let mut t = ArrayString::<2>::new();
     assert!(t.try_push_str(text).is_err());
     assert_eq!(&t, "");
 
@@ -326,9 +522,9 @@ fn test_string() {
     assert_eq!(tmut, "ab");
 
     // Test Error trait / try
-    let t = || -> Result<(), Box<Error>> {
-        let mut t = ArrayString::<[_; 2]>::new();
-        try!(t.try_push_str(text));
+    let t = || -> Result<(), Box<dyn Error>> {
+        let mut t = ArrayString::<2>::new();
+        t.try_push_str(text)?;
         Ok(())
     }();
     assert!(t.is_err());
@@ -338,7 +534,15 @@ fn test_string() {
 fn test_string_from() {
     let text = "hello world";
 	// Test `from` constructor
-    let u = ArrayString::<[_; 11]>::from(text).unwrap();
+    let u = ArrayString::<11>::from(text).unwrap();
+    assert_eq!(&u, text);
+    assert_eq!(u.len(), text.len());
+}
+
+#[test]
+fn test_string_parse_from_str() {
+    let text = "hello world";
+    let u: ArrayString<11> = text.parse().unwrap();
     assert_eq!(&u, text);
     assert_eq!(u.len(), text.len());
 }
@@ -354,9 +558,9 @@ fn test_string_from_bytes() {
 #[test]
 fn test_string_clone() {
     let text = "hi";
-    let mut s = ArrayString::<[_; 4]>::new();
+    let mut s = ArrayString::<4>::new();
     s.push_str("abcd");
-    let t = ArrayString::<[_; 4]>::from(text).unwrap();
+    let t = ArrayString::<4>::from(text).unwrap();
     s.clone_from(&t);
     assert_eq!(&t, &s);
 }
@@ -364,7 +568,7 @@ fn test_string_clone() {
 #[test]
 fn test_string_push() {
     let text = "abcαβγ";
-    let mut s = ArrayString::<[_; 8]>::new();
+    let mut s = ArrayString::<8>::new();
     for c in text.chars() {
         if let Err(_) = s.try_push(c) {
             break;
@@ -379,7 +583,7 @@ fn test_string_push() {
 
 #[test]
 fn test_insert_at_length() {
-    let mut v = ArrayVec::<[_; 8]>::new();
+    let mut v = ArrayVec::<_,  8>::new();
     let result1 = v.try_insert(0, "a");
     let result2 = v.try_insert(1, "b");
     assert!(result1.is_ok() && result2.is_ok());
@@ -389,7 +593,7 @@ fn test_insert_at_length() {
 #[should_panic]
 #[test]
 fn test_insert_out_of_bounds() {
-    let mut v = ArrayVec::<[_; 8]>::new();
+    let mut v = ArrayVec::<_,  8>::new();
     let _ = v.try_insert(1, "test");
 }
 
@@ -422,7 +626,7 @@ fn test_drop_in_insert() {
     flag.set(0);
 
     {
-        let mut array = ArrayVec::<[_; 2]>::new();
+        let mut array = ArrayVec::<_,  2>::new();
         array.push(Bump(flag));
         array.insert(0, Bump(flag));
         assert_eq!(flag.get(), 0);
@@ -437,7 +641,7 @@ fn test_drop_in_insert() {
 
 #[test]
 fn test_pop_at() {
-    let mut v = ArrayVec::<[String; 4]>::new();
+    let mut v = ArrayVec::<String,  4>::new();
     let s = String::from;
     v.push(s("a"));
     v.push(s("b"));
@@ -456,3 +660,70 @@ fn test_sizes() {
     let v = ArrayVec::from([0u8; 1 << 16]);
     assert_eq!(vec![0u8; v.len()], &v[..]);
 }
+
+#[test]
+fn test_default() {
+    use std::net;
+    let s: ArrayString<4> = Default::default();
+    // Something without `Default` implementation.
+    let v: ArrayVec<net::TcpStream,  4> = Default::default();
+    assert_eq!(s.len(), 0);
+    assert_eq!(v.len(), 0);
+}
+
+#[cfg(feature="array-sizes-33-128")]
+#[test]
+fn test_sizes_33_128() {
+    ArrayVec::from([0u8; 52]);
+    ArrayVec::from([0u8; 127]);
+}
+
+#[cfg(feature="array-sizes-129-255")]
+#[test]
+fn test_sizes_129_255() {
+    ArrayVec::from([0u8; 237]);
+    ArrayVec::from([0u8; 255]);
+}
+
+#[test]
+fn test_extend_zst() {
+    let mut range = 0..10;
+    #[derive(Copy, Clone, PartialEq, Debug)]
+    struct Z; // Zero sized type
+
+    let mut array: ArrayVec<_,  5> = range.by_ref().take(5).map(|_| Z).collect();
+    assert_eq!(&array[..], &[Z; 5]);
+    assert_eq!(range.next(), Some(5));
+
+    array.extend(range.by_ref().take(0).map(|_| Z));
+    assert_eq!(range.next(), Some(6));
+
+    let mut array: ArrayVec<_,  10> = (0..3).map(|_| Z).collect();
+    assert_eq!(&array[..], &[Z; 3]);
+    array.extend((3..5).map(|_| Z));
+    assert_eq!(&array[..], &[Z; 5]);
+    assert_eq!(array.len(), 5);
+}
+
+#[test]
+fn test_try_from_argument() {
+    use core::convert::TryFrom;
+    let v = ArrayString::<16>::try_from(format_args!("Hello {}", 123)).unwrap();
+    assert_eq!(&v, "Hello 123");
+}
+
+#[test]
+fn allow_max_capacity_arrayvec_type() {
+    // this type is allowed to be used (but can't be constructed)
+    let _v: ArrayVec<(), {usize::MAX}>;
+}
+
+#[should_panic(expected="ArrayVec: largest supported")]
+#[test]
+fn deny_max_capacity_arrayvec_value() {
+    if mem::size_of::<usize>() <= mem::size_of::<u32>() {
+        panic!("This test does not work on this platform. 'ArrayVec: largest supported'");
+    }
+    // this type is allowed to be used (but can't be constructed)
+    let _v: ArrayVec<(), {usize::MAX}> = ArrayVec::new();
+}
